use 5.006;
use strict;
use warnings FATAL => 'all';
use Module::Build;

my $builder = Module::Build->new(
                                 dist_name          => 'Linux::DesktopFiles',
                                 license            => 'perl',
                                 dist_author        => q{Șuteu "Trizen" Daniel <trizenx@gmail.com>},
                                 dist_version_from  => 'lib/Linux/DesktopFiles.pm',
                                 configure_requires => {
                                                        'Module::Build' => 0,
                                                       },
                                 requires => {
                                              'perl'      => 5.010,
                                              'GDBM_File' => 0,
                                             },
                                 add_to_cleanup     => ['Linux-DesktopFiles-*'],
                                 create_makefile_pl => 'traditional',
                                );

$builder->create_build_script();
